/* App layout */
html {
    overflow: scroll;
    overflow-x: hidden;
}

::-webkit-scrollbar {
    width: 0;  /* Remove scrollbar space */
    background: transparent;  /* Optional: just make scrollbar invisible */
}

::-webkit-scrollbar-thumb {
    background: #FF0000;
}

.app-layout {
    font-family: sans-serif;
    line-height: 20px;
    height: 100vh;
    /* border: black 1px solid; */
}

/* Tabs */
.instrument-tabs {
    margin-top: 2%;
    /* margin-left: 7%;
    margin-right: 7%; */
}

/* Header */
.header h1 {
    font-size: 40px;
    padding-top: 10px;
    padding-bottom: 10px;
    font-weight: 500;
}

/* Tables and table container */
.table-container {
    float: left;
    margin-right: 2%;
    width: 100%;
    height: calc(100vh - 175px);
    overflow-y: scroll;
}

.dash-spreadsheet.dash-freeze-top, .dash-spreadsheet.dash-virtualized {
    max-height: inherit !important;
}

.dash-table-container {
    margin-top: 30px;
    max-height: calc(100vh - 15px);
}

.run-button {
    line-height: 1.75;
    margin-top: 15px;
}

/* Plots and plot container */
.all-plots-container {
    float: left;
    width: 100%;
    height: calc(100vh - 160px);
    overflow-y: scroll;
}

.istd-plot-div {
    margin-top: 30px;
}

.istd-plot-div, .bio-plot-div {
    float: left;
    width: 100%;
}

.plot-container {
    margin-bottom: 30px;
    margin-left: 1%;
    margin-right: 1%;
    width: 98%;
}

h2 {
    text-align: center;
    margin: 40px;
}

.navbar-button {
    font-size: 1.15rem;
}

.istd-dropdown-style {
    display: inline-block;
    vertical-align: middle;
    width: 88%;
    margin-bottom: 5px;
    margin-right: 1%;
}

.istd-button-style {
    display: inline-block;
    vertical-align: middle;
    width: 11%;
    margin-bottom: 5px;
}

.istd-button-style button {
    margin-left: 1%;
    margin-right: 1%;
}

/* Radio items */
.radio-group-container {
    width: 100%;
    text-align: center;
}

.radio-group .form-check {
    padding-left: 0;
    display: inline-block;
}

.radio-group .btn-group > .form-check:not(:last-child) > .btn {
    border-top-right-radius: 0;
    border-bottom-right-radius: 0;
}

.radio-group .btn-group > .form-check:not(:first-child) > .btn {
    border-top-left-radius: 0;
    border-bottom-left-radius: 0;
    margin-left: -1px;
}

.parent-container {
    margin: auto;
    width: 100%;
}

/* Lazy CSS styles */
.margin-top-15 {
    margin-top: 15px;
}

.margin-top-30 {
    margin-top: 30px;
}

.checkbox-margin {
    margin-top: 1rem;
    margin-bottom: 0.5rem;
}

/* Modal styles */
.child-container {
    float: left;
    width: 50%;
}

.modal-styles {
    padding-left: 1.5vw;
    padding-right: 1.5vw;
}

.modal-styles-2 {
    padding-top: 3vh;
    padding-bottom: 3vh;
    padding-left: 3vw;
    padding-right: 3vw;
}

.modal-styles-3 {
    padding-left: 1.5vw;
    padding-right: 1.5vw;
    padding-top: 1vh;
    padding-bottom: 1vh;
}

.modal-styles-fullscreen {
    padding-left: 25vw;
    padding-right: 25vw;
    padding-bottom: 7.5vh;
}

.input-group-text .form-label {
    margin-bottom: 0;
}